import pytest
from pytest import approx
import numpy as np
import pandas as pd

import process_improve.univariate.metrics as univariate


class test_t_values:
    """
    Checks the calculation of t values (at a given 'alpha' and with an integer number of degrees
    of freedom), against the values from R.
    """

    assert univariate.t_value(0, 1) == np.NINF
    assert univariate.t_value(1, 2) == np.inf
    assert univariate.t_value(0.5, 3) == approx(0, rel=1e-16)

    # Tested in R:  qt(0.9, 5) ->  1.475884
    assert univariate.t_value(0.9, 5) == approx(1.475884, rel=1e-6)


class test_t_values_cdf:
    """
    Checks the calculation of t values (at a given 'alpha' and with an integer number of degrees
    of freedom), against the values from R.
    """

    assert univariate.t_value_cdf(0, 1) == 0.5
    assert univariate.t_value_cdf(np.NINF, 2) == 0
    assert univariate.t_value_cdf(np.inf, 3) == 1

    # Tested in R:  pt(0.9, 5) ->  0.7953144
    assert univariate.t_value_cdf(0.9, 5) == approx(0.7953144, rel=1e-8)
    # Tested in R:  pt(0.5, 1) ->  0.6475836
    assert univariate.t_value_cdf(0.5, 1) == approx(0.6475836, rel=1e-7)


def test_normality_check():

    """
    Tests on data actually from a normal distribution, and some data which is from a
    uniform distribution.

    In R:  version 3.6.0 (2019-04-26)
    > set.seed(42)
    > x = rnorm(10)
    > shapiro.test(x)
    > [1]  1.37095845 -0.56469817  ...
    > Shapiro-Wilk normality test
    > W = 0.9287, p-value = 0.4352

    > y = runif(10)
    > shapiro.test(y)
    > [1] 0.90403139 0.13871017 ...
    > Shapiro-Wilk normality test
    > W = 0.87415, p-value = 0.1117
    """
    x = [
        1.37095845,
        -0.56469817,
        0.36312841,
        0.63286260,
        0.40426832,
        -0.10612452,
        1.51152200,
        -0.09465904,
        2.01842371,
        -0.06271410,
    ]
    y = [
        0.90403139,
        0.13871017,
        0.98889173,
        0.94666823,
        0.08243756,
        0.51421178,
        0.39020347,
        0.90573813,
        0.44696963,
        0.83600426,
    ]

    # Data actually are from a  normal distribution:
    assert univariate.normality_check(x) == approx(0.4352, abs=1e-3)

    # Data actually are from a uniform distribution:
    assert univariate.normality_check(y) == approx(0.1117, abs=1e-3)


def test_univariate_robust_scale():
    """
    A scale estimator which is robust to outliers

    Testing against R code [R version 3.6.0 (2019-04-26)]
    > library(robustbase)
    # All code has the default argument: "finite.corr=TRUE"
    > robustbase::Sn(c(0, 1))                   # 0.8861018: FAILS in our implementation
    > robustbase::Sn(c(0, 1, 2))                # 2.207503
    > robustbase::Sn(c(0, 1, 2, 3))             # 1.13774: FAILS in our implementation
    > robustbase::Sn(c(0, 1, 2, 3, 4))          # 1.611203
    > robustbase::Sn(c(0, 1, 2, 3, 4, 5))       # 2.368504: FAILS in our implementation
    > robustbase::Sn(c(0, 1, 2, 3, 4, 5, 6))    # 2.85747
    > robustbase::Sn(c(0, 1, 2, 3, 4, 50, 6))   # 2.85747
    > robustbase::Sn(c(0, 1, 20, 3, 4, 50, 6))  # 5.714939: FAILS in our implementation
    > robustbase::Sn(c(0, 10, 20, 3, 4, 50, 6)) # 8.572409
    > robustbase::Sn(seq(1, 10))                # 3.5778: FAILS in our implementation
    > robustbase::Sn(seq(1, 11))                # 3.896614
    > robustbase::Sn(seq(1, 19))                # 6.259503
    > robustbase::Sn(seq(1, 1500))              # 447.225

    TODO: found this weird sequence that gives Sn of zero, even though there is variability:
    99, 95, 95, 100, 100, 100, 100, 95, 100, 100, 100, 100, 105, 105, 100, 95, 105, 100, 95, 100
    How to make it robust to this weird situation?
    """

    # Tests with an even number of samples, and small sample sizes do not agree with R.
    # This is because the R implementation aims for efficiency of calculation, and does not
    # follow the formula presented in the original paper.
    # Since we aim to be using this on medium/larger data sets, it should not matter.

    assert univariate.Sn(list(range(3))) == approx(2.207503, rel=1e-6)
    assert univariate.Sn(list(range(5))) == approx(1.611203, rel=1e-6)
    assert univariate.Sn(list(range(7))) == approx(2.85747, rel=1e-6)
    assert univariate.Sn([0, 10, 20, 3, 4, 50, 6]) == approx(8.572409, rel=1e-7)
    assert univariate.Sn(list(range(1, 12))) == approx(3.896614, rel=1e-6)
    assert univariate.Sn(list(range(1, 19))) == approx(5.3667, rel=1e-6)
    assert univariate.Sn(list(range(1, 20))) == approx(6.259503, rel=1e-6)
    # Corner cases:
    assert np.isnan(univariate.Sn([]))
    assert univariate.Sn([13]) == 0.0


def test_summary_stats_corner_case_with_robust_scale():
    x = [
        99,
        95,
        95,
        100,
        100,
        100,
        100,
        95,
        100,
        100,
        100,
        100,
        105,
        105,
        100,
        95,
        105,
        100,
        95,
        100,
    ]
    out = univariate.summary_stats(np.array(x), method="robust")
    assert out["center"] == np.mean(x)
    assert out["center"] != np.median(x)

    out = univariate.summary_stats(np.array(x), method="something-else")
    assert out["center"] == np.mean(x)
    assert out["center"] != np.median(x)


def test_median_abs_deviation():

    x = np.array([[10, 7, 4], [3, 2, 1]])
    assert univariate.median_abs_deviation(x, scale=1) == approx([3.5, 2.5, 1.5])
    assert univariate.median_abs_deviation(x, axis=None, scale=1) == 2.0

    from scipy import stats

    x = stats.norm.rvs(size=1000000, scale=2, random_state=123456)
    assert univariate.median_abs_deviation(x, scale=1) == approx(
        1.3487398527041636, rel=1e-12
    )
    assert univariate.median_abs_deviation(x) == approx(1.9996446978061115, rel=1e-12)

    with pytest.raises(TypeError, match=r"The argument 'center' must .*"):
        _ = univariate.median_abs_deviation(x, center=0.0)

    with pytest.raises(ValueError, match=r".* is not a valid scale value."):
        _ = univariate.median_abs_deviation(x, scale="robust")

    with pytest.raises(ValueError, match=r"nan_policy must be one of .*"):
        univariate.median_abs_deviation([1, 2, 3, 4], nan_policy="propogatess")

    with pytest.raises(TypeError):
        univariate.median_abs_deviation(["a", "b"])

    with pytest.raises(ValueError, match="The input contains nan values"):
        univariate.median_abs_deviation([np.nan, 1], nan_policy="raise")

    assert np.isnan(
        univariate.median_abs_deviation([np.nan, 1, 2], nan_policy="propagate")
    )

    assert np.isnan(
        univariate.median_abs_deviation(
            [
                np.nan,
            ],
            axis=None,
        )
    )
    assert np.isnan(
        univariate.median_abs_deviation(
            [
                np.nan,
            ],
            axis=0,
        )
    )
    assert np.isnan(univariate.median_abs_deviation([], axis=None))
    assert np.isnan(univariate.median_abs_deviation([], axis=0))
    assert np.isnan(
        univariate.median_abs_deviation(np.empty((2, 3, 4)) * np.nan, axis=None)
    )
    assert np.isnan(univariate.median_abs_deviation(np.array([np.nan, np.nan]), axis=0))


def test_t_test_differences():
    """
    Tests for the t-test of differences.

    R code to validate against:
    > sam = c(8.80, 6.60, 7.26, 9.32, 5.88, 8.44, 11.39, 6.82, 9.32, 5.63, 9.65, 9.49)
    > jen = c(5.37, 4.83, 7.87, 3.30, 8.26, 7.87, 8.26, 6.13, 6.13, 5.63, 2.96, 5.88)
    > t.test(sam, jen, paired = FALSE, var.equal = TRUE)
        data:  sam and jen
        t = 2.9906, df = 22, p-value = 0.00674
        alternative hypothesis: true difference in means is not equal to 0
        95 percent confidence interval:
        0.6669748 3.6846919
        sample estimates:
        mean of x mean of y
        8.216667  6.040833
    """
    sam = [8.80, 6.60, 7.26, 9.32, 5.88, 8.44, 11.39, 6.82, 9.32, 5.63, 9.65, 9.49]
    jen = [5.37, 4.83, 7.87, 3.30, 8.26, 7.87, 8.26, 6.13, 6.13, 5.63, 2.96, 5.88]
    mik = [5.80, 9.00, 5.60, 8.40, 8.60, None, None, None, None, None, None, None]
    temp = pd.DataFrame(data={"Sam": sam, "Jen": jen, "Mik": mik})
    temp.reset_index(inplace=True)
    temp = temp.melt(id_vars="index")
    df = temp.drop("index", axis=1).dropna().rename(columns={"variable": "Person"})
    output = univariate.ttest_difference(
        df, grouper_column="Person", values_column="value", conflevel=0.95
    )
    row = output[output["Group A name"].eq("Sam") & output["Group B name"].eq("Jen")]

    # Assert against the R values in the above validation script.
    # Note: in R, the test for t.test(A, B), checking A minus B.
    # We define our test as group B minus group A. Therefore we have to flip our signs,
    # and high/low values of the confidence interval
    assert row["Group A average"][0] == approx(8.216667, rel=1e-5)
    assert row["Group B average"][0] == approx(6.040833, rel=1e-5)
    assert row["z value"][0] == approx(-2.9906, rel=1e-4)
    assert row["p value"][0] == approx(0.00674, rel=1e-3)
    assert row["ConfInt: Lo"][0] == approx(-3.6846919, rel=1e-4)
    assert row["ConfInt: Hi"][0] == approx(-0.6669748, rel=1e-4)
    assert row["Degrees of freedom"][0] == approx(22, rel=1e-8)


def test_t_paried_test_differences():
    """
    Tests for the paired t-test of differences.

    R code to validate against:
    > sam = c(8.80, 6.60, 7.26, 9.32, 5.88, 8.44, 11.39, 6.82, 9.32, 5.63, 9.65, 9.49)
    > jen = c(5.37, 4.83, 7.87, 3.30, 8.26, 7.87, 8.26, 6.13, 6.13, 5.63, 2.96, 5.88)
    > t.test(sam, jen, paired = FALSE, var.equal = TRUE)
            Paired t-test

        data:  sam and jen
        t = 2.8139, df = 11, p-value = 0.01685
        alternative hypothesis: true difference in means is not equal to 0
        95 percent confidence interval:
        0.4739104 3.8777563
        sample estimates:
        mean of the differences
                    2.175833
    """
    sam = [8.80, 6.60, 7.26, 9.32, 5.88, 8.44, 11.39, 6.82, 9.32, 5.63, 9.65, 9.49]
    jen = [5.37, 4.83, 7.87, 3.30, 8.26, 7.87, 8.26, 6.13, 6.13, 5.63, 2.96, 5.88]
    temp = pd.DataFrame(data={"Sam": sam, "Jen": jen})
    temp.reset_index(inplace=True)
    temp = temp.melt(id_vars="index")
    df = temp.drop("index", axis=1).dropna().rename(columns={"variable": "Person"})
    output = univariate.ttest_paired_difference(
        df, grouper_column="Person", values_column="value", conflevel=0.95
    )
    row = output[output["Group A name"].eq("Sam") & output["Group B name"].eq("Jen")]

    # Assert against the R values in the above validation script.
    # Note: in R, the test for t.test(A, B), checking A minus B.
    assert row["Group A average"][0] == approx(8.216667, rel=1e-7)
    assert row["Group B average"][0] == approx(6.040833, rel=1e-7)
    assert row["Differences mean"][0] == approx(2.175833, rel=1e-6)
    assert row["z value"][0] == approx(2.8139, rel=1e-4)
    assert row["p value"][0] == approx(0.01685, rel=1e-4)
    assert row["ConfInt: Lo"][0] == approx(0.4739104, rel=1e-7)
    assert row["ConfInt: Hi"][0] == approx(3.8777563, rel=1e-7)
    assert row["Degrees of freedom"][0] == 11


@pytest.fixture
def univariate_summary():
    """
    A univariate case study

    In R:
    r <- c(108, 89.52, 95.16, 101.61, 99.19, 100, 93.55, 97.58, 93.55, 98.39, 88.71)

    r_mean <- mean(r)                           # 96.8418181818182
    r_std_ddof1 <- sd(r)                        # 5.56692521627841
    r_rsd <- r_std_ddof1 / r_mean               # 0.05748472427301
    r_median <- median(r)                       # 97.58
    r_iqr <- IQR(r)                             # 6.045
    r_Sn <- robustbase::Sn(r)                   # 6.28653702970298
    r_rsd_robust <- r_Sn / r_median             # 0.06442444178831

    r_min <- min(r)                             # 88.71
    r_max <- max(r)                             # 108
    r_n <- length(r)                            # 11

    r_percentile_5 <- quantile(r, probs = 0.05) # 89.115
    r_percentile_25<- quantile(r, probs = 0.25) # 93.55
    r_percentile_75<- quantile(r, probs = 0.75) # 99.595
    r_percentile_95<- quantile(r, probs = 0.95) # 104.805


    See the tests for "Test__univariate_robust_scale", to understand why an odd number of
    samples were chosen.
    """
    y = [108.0, 89.52, 95.16, 101.61, 99.19, 100, 93.55, 97.58, 93.55, 98.39, 88.71]
    return pd.DataFrame(data={"values": y})


def test_compare_to_R_with_without_missing(univariate_summary):
    # Verifies that we can reproduce results from R. R version 3.6.0 (2019-04-26)
    # Checked on 21 February 2020.
    data = univariate_summary
    for k in range(2):
        if k > 0:
            # For the second loop: add a missing value and ensure you get the same results
            # as for the first loop (without missing values)
            data = data.append({"values": np.nan}, ignore_index=True)

        out = univariate.summary_stats(data["values"])
        assert out["mean"] == approx(96.84181818181816, abs=1e-8)
        assert out["std_ddof1"] == approx(5.566925216278406, abs=1e-8)
        assert out["rsd_classical"] == approx(0.05748472427301, abs=1e-8)
        assert out["median"] == approx(97.58, abs=1e-8)
        assert out["center"] == approx(97.58, abs=1e-8)  # center = median by default
        assert out["iqr"] == approx(6.045, abs=1e-8)
        assert out["spread"] == approx(
            6.28653702970298, abs=1e-8
        )  # spread = Sn (changed in 0.5)
        assert out["rsd"] == approx(0.06442444178831, abs=1e-8)
        assert out["min"] == 88.71
        assert out["max"] == 108
        assert out["N_non_missing"] == 11

        # Note: there are differences in how Numpy and R interpolate the results
        assert out["percentile_05"] == approx(89.115, rel=1e-6)
        assert out["percentile_25"] == approx(93.55, rel=1e-6)
        assert out["percentile_75"] == approx(99.595, rel=1e-6)
        assert out["percentile_95"] == approx(104.805, rel=1e-6)


def test_as_numpy_array(univariate_summary):
    out = univariate.summary_stats(univariate_summary["values"].values)
    assert out["mean"] == approx(96.84181818181816, abs=1e-8)
    assert out["std_ddof1"] == approx(5.566925216278406, abs=1e-8)


def test__raises_error():
    with pytest.raises(ValueError, match="Expecting a NumPy vector or Pandas series."):
        univariate.summary_stats([1, 2, 3, 3, 2, 1])


def test_confidence_interval():
    """
    Test confidence intervals.
    r1 <- c(108, 89.52, 95.16, 101.61, 99.19, 100, 93.55, 97.58, 93.55, 98.39, 88.71, 94.35)

    Results of the CI, compared to R.
    t.test(r1-90)
    """
    y = [
        108.0,
        89.52,
        95.16,
        101.61,
        99.19,
        100,
        93.55,
        97.58,
        93.55,
        98.39,
        88.71,
        94.35,
    ]
    data = pd.DataFrame(data={"values": y})
    expected_LB = 3.230888
    expected_UB = 10.037445

    out = univariate.confidence_interval(
        data - 90, "values", conflevel=0.95, style="regular"
    )
    assert out[0] == approx(expected_LB, abs=1e-4)
    assert out[1] == approx(expected_UB, abs=1e-4)
    out = univariate.confidence_interval(
        data - 90, "values", conflevel=0.95, style="robust"
    )
    # TODO: complete the test for the robust case


@pytest.fixture
def within_between_sd_data():

    """
    r1 <- c(108.06, 89.52, 95.16, 101.61, 99.19, 100, 93.55, 97.58, 93.55, 98.39, 96.77, 89.92,
            88.71, 94.35)
    r2 <- c(108.07, 87.9, 95.97, 97.58, 100, 95.97, 88.71, 97.59, 95.97, 93.55, 96.78, 86.69,
            87.1, 93.55)
    """
    replicate1 = [
        108.06,
        89.52,
        95.16,
        101.61,
        99.19,
        100,
        93.55,
        97.58,
        93.55,
        98.39,
        96.77,
        89.92,
        88.71,
        94.35,
    ]
    replicate2 = [
        108.07,
        87.9,
        95.97,
        97.58,
        100,
        95.97,
        88.71,
        97.59,
        95.97,
        93.55,
        96.78,
        86.69,
        87.1,
        93.55,
    ]
    temp = pd.DataFrame(data={"e1": replicate1, "e2": replicate2})
    temp.reset_index(inplace=True)
    temp = temp.melt(id_vars="index")
    df = temp.drop("variable", axis=1)
    empty = pd.DataFrame(columns=["value", "index"])
    return df, empty


def test_within_between_variance(within_between_sd_data):

    """
    Results are from a spreadsheet template. Unsure of the origin, or accuracy.
    """
    df, _ = within_between_sd_data
    expected_within_ms = 1.916015 ** 2
    expected_between_sd = 7.659146 ** 2
    expected_actual_sd = 5.490761 ** 2
    dof_within = 14
    dof_between = 13
    dof_total = 27

    out = univariate.within_between_standard_deviation(df, "value", "index")
    assert out["total_ms"] == approx(expected_actual_sd, rel=1e-5)
    assert out["total_dof"] == dof_total
    assert out["within_ms"] == approx(expected_within_ms, rel=1e-5)
    assert out["within_dof"] == dof_within
    assert out["between_ms"] == approx(expected_between_sd, rel=1e-5)
    assert out["between_dof"] == dof_between


def test_empty_case(within_between_sd_data):
    """
    What happens if there are no data? Everything should be zero.
    """
    _, empty = within_between_sd_data
    out = univariate.within_between_standard_deviation(empty, "value", "index")
    assert out["total_ms"] == 0
    assert out["within_ms"] == 0
    assert out["within_dof"] == 0
    assert out["between_ms"] == 0
    assert out["total_dof"] == 0
    assert out["between_dof"] == 0


def test_within_between_sd_missing_values():
    """
    Test against Excel sheet formulas.
    r1 <- c(108.06, NA, 95.16, 101.61, 99.19, 100, 93.55, 97.58, 93.55, 98.39, 96.77, 89.92,
            88.71, NA)
    r2 <- c(108.07, 87.9, 95.97, 97.58, 100, 95.97, 88.71, 97.59, 95.97, 93.55, 96.78, 86.69,
            87.1, NA)
    """

    empty = pd.DataFrame(columns=["value", "index"])
    empty.append({"index": 1, "value": np.NaN}, ignore_index=True)
    empty.append({"index": 1, "value": 123456}, ignore_index=True)

    replicate1 = [
        108.06,
        np.nan,
        95.16,
        101.61,
        99.19,
        100,
        93.55,
        97.58,
        93.55,
        98.39,
        96.77,
        89.92,
        88.71,
        np.nan,
    ]
    replicate2 = [
        108.07,
        87.9,
        95.97,
        97.58,
        100,
        95.97,
        88.71,
        97.59,
        95.97,
        93.55,
        96.78,
        86.69,
        87.1,
        np.nan,
    ]
    temp = pd.DataFrame(data={"e1": replicate1, "e2": replicate2})
    temp.reset_index(inplace=True)
    temp = temp.melt(id_vars="index")
    df = temp.drop("variable", axis=1)

    # within_between_variance_missing_values
    # Results are from a spreadsheet template. Unsure of the origin, or accuracy.
    expected_within_ms = 2.036406 ** 2
    expected_between_sd = 7.754816 ** 2
    expected_actual_sd = 5.669397 ** 2
    dof_within = 12
    dof_between = 12
    dof_total = 24

    out = univariate.within_between_standard_deviation(df, "value", "index")
    assert out["total_ms"] == approx(expected_actual_sd, abs=1e-4)
    assert out["total_dof"] == dof_total
    assert out["within_ms"] == approx(expected_within_ms, abs=1e-4)
    assert out["within_dof"] == dof_within
    assert out["between_ms"] == approx(expected_between_sd, abs=1e-4)
    assert out["between_dof"] == dof_between

    # test__empty_case
    # What happens if there are no/little data after accounting for outliers?
    out = univariate.within_between_standard_deviation(empty, "value", "index")
    assert out["total_ms"] == 0
    assert out["within_ms"] == 0
    assert out["within_dof"] == 0
    assert out["between_ms"] == 0
    assert out["total_dof"] == 0
    assert out["between_dof"] == 0


@pytest.fixture
def outliers_data():

    # Rosner data set: https://www.itl.nist.gov/div898/handbook/eda/section3/eda35h3.htm
    rosner = [
        -0.25,
        0.68,
        0.94,
        1.15,
        1.20,
        1.26,
        1.26,
        1.34,
        1.38,
        1.43,
        1.49,
        1.49,
        1.55,
        1.56,
        1.58,
        1.65,
        1.69,
        1.70,
        1.76,
        1.77,
        1.81,
        1.91,
        1.94,
        1.96,
        1.99,
        2.06,
        2.09,
        2.10,
        2.14,
        2.15,
        2.23,
        2.24,
        2.26,
        2.35,
        2.37,
        2.40,
        2.47,
        2.54,
        2.62,
        2.64,
        2.90,
        2.92,
        2.92,
        2.93,
        3.21,
        3.26,
        3.30,
        3.59,
        3.68,
        4.30,
        4.64,
        5.34,
        5.42,
        6.01,
    ]

    sequence = [
        9101,
        9193,
        9440,
        9836,
        9677,
        9515,
        9783,
        9130,
        9469,
        9528,
        np.nan,
        np.nan,
        9805,
        9894,
        9941,
        10140,
        9001,
        9178,
        10080,
        9816,
        9160,
        8862,
        9376,
        9515,
        10670,
        10090,
        9979,
        9761,
        9422,
        9696,
        10130,
        10090,
        9641,
        9771,
        9503,
        9533,
        9413,
        9194,
        9219,
        9756,
        np.nan,
        np.nan,
        9240,
        9337,
        9682,
        9809,
        9343,
        9366,
        9245,
        9190,
        9363,
        9273,
        9500,
        9550,
        9664,
        9320,
        9247,
        9095,
        9122,
        9272,
        9157,
        9100,
        10670,
        10900,
        8899,
        8838,
        9203,
        9403,
        9520,
        9123,
        9109,
        9857,
        9936,
        9312,
        9225,
    ]
    return rosner, sequence


def test_rosner_nonrobust_esd(outliers_data):
    rosner, _ = outliers_data
    outliers, reasons = univariate.outlier_detection_multiple(
        rosner,
        algorithm="esd",
        max_outliers_detected=7,
        robust_variant=False,
        alpha=0.05,
    )

    # Ensure the vector is unchanged afterwards
    assert len(rosner) == 54
    assert rosner[0] == -0.25
    assert rosner[-1] == 6.01
    assert outliers == [53, 52, 51]

    # Compare values in the explanation from NIST:
    # https://www.itl.nist.gov/div898/handbook/eda/section3/eda35h3.htm
    assert reasons["lambda"] == approx(
        [3.158, 3.151, 3.143, 3.136, 3.128, 3.120, 3.111], rel=1e-3
    )
    assert reasons["R_i"] == approx(
        [3.118, 2.942, 3.179, 2.810, 2.815, 2.848, 2.279], rel=1e-3
    )


def test_rosner_esd_kwargs(outliers_data):
    """
    In this example it picks up fewer outliers.
    """
    rosner, _ = outliers_data
    outliers, _ = univariate.outlier_detection_multiple(
        rosner,
        algorithm="esd",
        max_outliers_detected=7,
        robust_variant=True,
        alpha=0.05,
    )
    assert outliers == [53]


def test_rosner_esd_no_outliers(outliers_data):
    """
    In this example it picks up no outliers. Ensures that the test can also return an empty
    list.
    """
    rosner, _ = outliers_data
    outliers, _ = univariate.outlier_detection_multiple(
        rosner[1:-5],
        algorithm="esd",
        max_outliers_detected=0,
    )
    assert outliers == []


def test_rosner_esd_corner_case():
    """
    In this example it picks up no outliers. Ensures that the test can also return an empty
    list.
    """
    outliers, extra_out = univariate.outlier_detection_multiple(
        [1, 2], algorithm="esd", max_outliers_detected=1
    )
    assert extra_out["p-value"][0] == 0

    # If the values are all the same:
    outliers, extra_out = univariate.outlier_detection_multiple(
        [3, 3, 3], algorithm="esd", max_outliers_detected=1
    )
    assert np.isnan(extra_out["p-value"])
    assert extra_out["cutoff"] == -1
    assert len(outliers) == 0

    outliers, extra_out = univariate.outlier_detection_multiple(
        [2, 2], algorithm="esd", max_outliers_detected=1
    )
    assert np.isnan(extra_out["p-value"])
    assert extra_out["cutoff"] == -1
    assert len(outliers) == 0

    outliers, extra_out = univariate.outlier_detection_multiple(
        [1], algorithm="esd", max_outliers_detected=1
    )
    assert np.isnan(extra_out["p-value"])
    assert extra_out["cutoff"] == -1
    assert len(outliers) == 0

    outliers, extra_out = univariate.outlier_detection_multiple(
        [1, 2, 3], algorithm="something-else", max_outliers_detected=1
    )
    assert len(extra_out) == 0
    assert len(outliers) == 0


def test_sequence_compare_R(outliers_data):
    """
    Compares it to an R sequence and the Grubb's test there.
    """
    _, sequence = outliers_data
    outliers, reasons_regular = univariate.outlier_detection_multiple(
        sequence,
        algorithm="esd",
        max_outliers_detected=1,
        robust_variant=False,
        alpha=0.05,
    )
    assert reasons_regular["p-value"][0] == approx(0.02066273, rel=1e-7)

    # Now with the robust version, to check NaN handling.
    outliers, reasons_robust = univariate.outlier_detection_multiple(
        sequence,
        algorithm="esd",
        max_outliers_detected=1,
        robust_variant=True,
        alpha=0.05,
    )
    assert outliers[0] == 63
    assert np.isnan(
        univariate.median_abs_deviation(
            [
                np.nan,
            ],
            axis=None,
        )
    )


def test_distribution_check():
    """
    R code for the KS test:

    > y1 = []
    > ks.test(y1,"pnorm")
    """
    # TODO
    pass



import pytest
import numpy as np
from process_improve.bivariate.methods import find_elbow_point


@pytest.fixture
def straight_line():
    slope = 2.5
    intercept = -13
    delta = 0.1
    x = np.arange(0, 5, delta)
    y0 = np.zeros_like(x)
    y1 = slope * x + intercept
    y2 = np.ones_like(x)
    return x, y0, y1, y2


def test__validate_nan_output(straight_line):
    """
    Is the elbow where we expect it? For clean, perfect data.
    """
    x, y0, y1, y2 = straight_line
    # Horizontal line
    elbow = find_elbow_point(x, y0)
    assert np.isnan(elbow)

    # Vertical line
    elbow = find_elbow_point(x, y2)
    assert np.isnan(elbow)

    # A line, but no elbow.
    elbow = find_elbow_point(x, y1)
    assert np.isnan(elbow)


@pytest.fixture
def elbow_with_synthetic_data():
    """
    Creates simulated data to test against.

    Create a line with slope of 2.0 between 0 and 5 on the x-axis.
    Then a line of slope of 3.0 between 5 and 10.
    The intersection is of course an elbow, at the break point of x=5.
    Add no noise: the function must still be able to find it.
    """
    delta = 0.1
    slope_2 = 2
    slope_3 = 3
    break_pt = 5
    intercept_2 = 0.0
    line_2 = np.arange(0, break_pt, delta) * slope_2 + intercept_2
    line_3 = (
        np.arange(break_pt, break_pt * 2, delta) * slope_3
        + (slope_2 - slope_3) * break_pt
    )
    x = np.arange(0, break_pt * 2, delta)
    y = np.concatenate((line_2, line_3))
    break_pt = break_pt

    return x, y, break_pt


def test__validate_with_synthetic_data(elbow_with_synthetic_data):
    """
    Is the elbow where we expect it? For clean, perfect data.
    """
    x, y, break_pt = elbow_with_synthetic_data
    expected_elbow = np.argmin(np.abs(x - break_pt))
    assert expected_elbow == find_elbow_point(x, y)


def test__validate_with_synthetic_data_plus_noise(elbow_with_synthetic_data):
    """
    Is the elbow where we expect it? For low noisy data. Use normally distributed
    noise, with standard deviation = 2% of the range of y.

    Elbow should be found within +/- 10 index positions from actual point
    """
    # Test with an odd number of points
    x, y, break_pt = elbow_with_synthetic_data
    np.random.seed(12)
    range = (y.max() - y.min()) * 0.02
    y = y + np.random.randn(y.shape[0]) * range
    expected_elbow = np.argmin(np.abs(x - break_pt))
    found_elbow = find_elbow_point(x, y)
    assert abs(found_elbow - expected_elbow) < 10

    # Test with an even number of points
    x, y, break_pt = elbow_with_synthetic_data
    range = (y.max() - y.min()) * 0.02
    y = y + np.random.randn(y.shape[0]) * range
    expected_elbow = np.argmin(np.abs(x - break_pt))
    found_elbow = find_elbow_point(x, y, max_iter=40)
    assert abs(found_elbow - expected_elbow) < 10


def test__corner_case():
    found_elbow = find_elbow_point(
        [np.nan] * 11,
        [np.nan] * 11,
    )
    assert found_elbow == -1



import warnings
from collections import defaultdict
from typing import Tuple, List, DefaultDict, Any

import numpy as np
import pandas as pd
from scipy.stats import shapiro, t

__eps = np.finfo(np.float32).eps


def t_value(p, v):
    r"""
    Returns the value on the x-axis if you plot the cumulative t-distribution with a fractional
    area of `p` (p is therefore a fractional value between 0 and 1 on the y-axis) and `v` is the
    degrees of freedom.


    Examples
    --------

    Since the cumulative distribution passes symmetrically through the x-axis at 0.0 for any
    number of degrees of freedom

    >>> t_value(0.5, v)
    0.0

    Zero fractional area under the curve is always at :math:`-\infty`:

    >>> t_value(0.0, v)
    -Inf

    100% fractional area is always at :math:`+\infty`:

    >>> t_value(1.0, v)
    +Inf

    See also
    --------
    t_value_cdf: does the inverse of this function.
    """
    return t.ppf(p, df=v)


def t_value_cdf(z, v):
    r"""
    Returns the value on the y-axis if you plot the cumulative t-distribution with a fractional
    area of `p` (p is therefore a fractional value between 0 and 1 on the y-axis) and `v` is the
    degrees of freedom.


    Examples
    --------

    Since the cumulative distribution passes symmetrically through the x-axis at 0.0 for any
    number of degrees of freedom

    >>> t_value_cdf(z=0.0, v)
    0.5

    Zero fractional area under the curve is always at :math:`-\infty`:

    >>> t_value_cdf(np.ninf, v)
    0.0

    100% fractional area is always at :math:`+\infty`:

    >>> t_value(np.inf, v)
    1.0

    See also
    --------
    t_value: does the inverse of this function.
    """
    return t.cdf(z, df=v)


def normality_check(x):
    """
    The p-value of the hypothesis that the data are from a normal distribution.

    If the p-value is less than the chosen alpha level (e.g. 0.05 or 0.025), then there is
    evidence that the data tested are NOT normally distributed.

    On the other hand, if the p-value is greater than the chosen alpha level, then the null
    hypothesis that the data came from a normally distributed population can not be rejected.
    NOTE: it does not mean that the data are normally distributed, just that we have nothing better
    to say about it. See the `Shapiro-Wilk test`_.

    Implementation

    Uses the Shapiro Wilk test directly taken from `scipy.stats.shapiro`_.

    .. _Shapiro-Wilk test:
        https://en.wikipedia.org/wiki/Shapiro-Wilk_test

    .. _scipy.stats.shapiro:
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.shapiro.html
    """
    output = shapiro(x)
    return output[1]


def Sn(x, constant=1.1926):
    """
    Computes a robust scale estimator. The Sn metric is an efficient alternative to MAD.

    Args:
        x (iterable): A vector of values

    Outputs:
        a scalar value, the Sn estimate of spread



    The `constant` gives values which are consistent with iid values from a Gaussian distribution
    and no outliers.

    Tested against once of the most reliable open-source packages, written by some of the
    most respected names in the area of robust methods: [1]_ and [2]_.

    Disadvantages of MAD:

    *   It does not have particularly high efficiency for data that is in fact normal (37%).
        In comparison, the median has 64% efficiency for normal data.

    *   The MAD statistic also has an implicit assumption of symmetry. That is, it measures the
        distance from a measure of central location (the median).


    References
    ----------

    .. [1] https://cran.r-project.org/web/packages/robustbase/
    .. [2] Rousseeuw, Peter J.; Croux, Christophe (December 1993),
        "Alternatives to the Median Absolute Deviation", Journal of the American Statistical
        Association, American Statistical Association, 88 (424): 1273–1283,
        https://dx.doi.org/10.2307/2291267

    """
    n = np.sum(~np.isnan(x))
    if n == 0:
        return np.float64(np.nan)
    elif n == 1:
        return np.float64(0.0)
    medians = []
    for i, value in enumerate(x):
        # In the paper by Rousseeuw and Croux: they seem to iterate over all data. But
        # in the R-source code, they iterate over the two loops for i/= j. That also makes
        # sense, because the difference with itself does not help determine the spread. The
        # spread is exactly the average (median) of the deviations.

        if np.isnan(value):
            continue

        differences = []
        for j, other in enumerate(x):
            if np.isnan(other):
                continue
            differences.append(abs(value - other))

        # Belongs to the outer loop
        medians.append(np.nanmedian(differences))

    if n <= 9:
        # Correction factors for n = 2 to 9:
        correction = [0.743, 1.851, 0.954, 1.351, 0.993, 1.198, 1.005, 1.131][n - 2]
    elif n % 2:
        correction = n / (n - 0.9)  # odd number of cases
    else:
        correction = 1.0

    return constant * np.nanmedian(medians) * correction


def _contains_nan(a, nan_policy="propagate"):
    """
    From scipy.stats.stats
    """
    policies = ["propagate", "raise", "omit"]
    if nan_policy not in policies:
        raise ValueError(
            "nan_policy must be one of {%s}" % ", ".join("'%s'" % s for s in policies)
        )
    try:
        # Calling np.sum to avoid creating a huge array into memory
        # e.g. np.isnan(a).any()
        with np.errstate(invalid="ignore"):
            contains_nan = np.isnan(np.sum(a))
    except TypeError:
        # This can happen when attempting to sum things which are not
        # numbers (e.g. as in the function `mode`). Try an alternative method:
        try:
            contains_nan = np.nan in set(a.ravel())
        except TypeError:  # pragma: no cover
            # Don't know what to do. Fall back to omitting nan values and
            # issue a warning.
            contains_nan = False
            nan_policy = "omit"
            warnings.warn(
                "The input array could not be properly checked for nan "
                "values. nan values will be ignored.",
                RuntimeWarning,
            )

    if contains_nan and nan_policy == "raise":
        raise ValueError("The input contains nan values")

    return (contains_nan, nan_policy)


def ttest_difference_calculate(sample_A, sample_B, conflevel=0.995) -> dict:
    """Core calculation for a test of differences between the average of A and the average of B.
    No checking of inputs.

    Parameters
    ----------
    sample_A : iterable
        Vector of n_a measurements.
    sample_B : iterable
        Vector of n_b measurements.
    conflevel : float
        Value between 0 and 1 (closer to 1.0), that gives the level of confidence required for
        the 2-sided test.

    Returns
    -------
    dict
        Outcomes from the statistical test.
    """
    axis = 0
    v1, v2 = sample_A.var(axis, ddof=1), sample_B.var(axis, ddof=1)
    n_A, n_B = sample_A.shape[axis], sample_B.shape[axis]
    m1, m2 = sample_A.mean(), sample_B.mean()
    df = n_A + n_B - 2.0
    ct = abs(t_value(p=(1 - conflevel) / 2.0, v=df))
    d = m2 - m1
    svar = ((n_A - 1) * v1 + (n_B - 1) * v2) / df

    with np.errstate(divide="ignore", invalid="ignore"):
        sd_z_variate = np.sqrt(svar * (np.divide(1.0, n_A) + np.divide(1.0, n_B)))
        z_variate = np.divide(d, sd_z_variate)

    confint_lo = d - ct * sd_z_variate
    confint_hi = d + ct * sd_z_variate

    return {
        "Group A number": int(n_A),
        "Group B number": int(n_B),
        "Group A average": sample_A.mean(),
        "Group B average": sample_B.mean(),
        "z value": z_variate,
        "ConfInt: Lo": confint_lo,
        "ConfInt: Hi": confint_hi,
        "p value": 2 * t_value_cdf(-np.abs(z_variate), df),
        "Degrees of freedom": df,
        "Pooled standard deviation": sd_z_variate,
    }


def ttest_difference(
    df: pd.DataFrame, grouper_column: str, values_column: str, conflevel=0.995
):
    """
    Calculates the t-test for differences between two or more groups and returns a confidence
    interval for the difference. The test is for UNPAIRED differences.

    The dataframe `df` contains a `grouper_column` with 2 or more unique values (e.g. 'A' and 'B').
    All unique values of the `grouper_column` are used, and t-tests are done between the values
    in the `values_column`.

    Args:
        df (pd.DataFrame): Dataframe of the values and grouping variable.
        grouper_column (str): Indicates which column will be grouped on.
        values_column (str): Which column contains the numeric values to calculate the test on.
        conflevel (float, optional): [description]. Defaults to 0.995.

    Output: Dataframe with columns containing the statistical outputs of the t-test, including:
        1. Group "A" name
        2. Group "B" name
        3. Group "A" mean
        4. Group "B" mean
        5. z-value for the difference between group "B" minus group "A"
        6. p-value for this z-value
        7. Confidence interval low value for difference between group "B" minus group "A"
        8. Confidence interval high value for difference between group "B" minus group "A"

    Example:
    df : has 3 levels in the grouper variable; ['Marco', 'Pete', 'Sam']

    Output will have 3 rows:
    Group A name  Group B name
    Marco         Pete
    Marco         Sam
    Pete          Sam
    """

    data_subset = df[[grouper_column, values_column]].copy()
    data_subset = data_subset.dropna()
    groups = df[grouper_column].unique()
    output = pd.DataFrame()
    groups = list(groups)
    while len(groups) > 0:
        groupA_name = groups.pop(0)
        for groupB_name in groups:

            sample_A = data_subset[data_subset[grouper_column].eq(groupA_name)][
                values_column
            ]
            sample_B = data_subset[data_subset[grouper_column].eq(groupB_name)][
                values_column
            ]
            sample_A = sample_A.astype(np.float64)
            sample_B = sample_B.astype(np.float64)
            basic_stats = ttest_difference_calculate(sample_A, sample_B, conflevel)
            basic_stats.update(
                {
                    "Group A name": groupA_name,
                    "Group B name": groupB_name,
                }
            )
            output = output.append(basic_stats, ignore_index=True)

    return output


def ttest_paired_difference_calculate(differences, conflevel=0.995) -> dict:
    """Core calculation for a test of differences.

    Parameters
    ----------
    sample_A : iterable
        Vector of n_a measurements.
    sample_B : iterable
        Vector of n_b measurements.
    conflevel : float
        Value between 0 and 1 (closer to 1.0), that gives the level of confidence required for
        the 2-sided test.

    Returns
    -------
    dict
        Outcomes from the statistical test.
    """
    diff_mean = differences.mean()
    diff_svar = differences.std(ddof=1)
    dof = differences.shape[0] - 1  # n-1 d

    # By the central limit theorem, the `differences` values should be normally distributed
    # with average (central value) given by mean of group A values, minus mean of group B values.
    # Scale factor is the standard deviations of `differences`: estimated, therefore t-distribution
    ct = abs(t_value(p=(1 - conflevel) / 2.0, v=dof))

    with np.errstate(divide="ignore", invalid="ignore"):
        sd_z_variate = diff_svar * np.sqrt(np.divide(1.0, dof + 1))
        z_variate = np.divide(diff_mean, sd_z_variate)

    confint_lo = diff_mean - ct * sd_z_variate
    confint_hi = diff_mean + ct * sd_z_variate

    return {
        "Differences mean": diff_mean,
        "z value": z_variate,
        "ConfInt: Lo": confint_lo,
        "ConfInt: Hi": confint_hi,
        "p value": 2 * t_value_cdf(-np.abs(z_variate), dof),
        "Degrees of freedom": dof,
        "Standard deviation": sd_z_variate,
    }


def ttest_paired_difference(
    df: pd.DataFrame, grouper_column: str, values_column: str, conflevel=0.995
):
    """
    Calculates the t-test for paired differences between two or more groups and returns a
    confidence interval for the difference. The test is for PAIRED differences.
    The differences is always defined as the A values minus the B values: after - before, or A - B.

    The dataframe `df` contains a `grouper_column` with 2 or more unique values (e.g. 'A' and 'B').
    All unique values of the `grouper_column` are used, and t-tests are done between the values
    in the `values_column`.

    When selecting the columns, the number of values per column must be the same.

    Args:
        df (pd.DataFrame): Dataframe of the values and grouping variable.
        grouper_column (str): Indicates which column will be grouped on.
        values_column (str): Which column contains the numeric values to calculate the test on.
        conflevel (float, optional): [description]. Defaults to 0.995.

    Output: Dataframe with columns containing the statistical outputs of the t-test, including:
        1. Group A name
        2. Group B name
        3. Group A mean
        4. Group B mean
        5. Differences mean:  The average difference between the groups
            (Note: this is not the same as the difference of the averages, given by 4 and 5 above!)
        CHECK5. z-value for the difference between group "B" minus group "A"
        CHECK6. p-value for this z-value
        CHECK7. Confidence interval low value for difference between group "B" minus group "A"
        CHECK8. Confidence interval high value for difference between group "B" minus group "A"

    """
    data_subset = df[[grouper_column, values_column]].copy()
    data_subset = data_subset.dropna()
    groups = df[grouper_column].unique()
    output = pd.DataFrame()
    groups = list(groups)
    while len(groups) > 0:
        groupA_name = groups.pop(0)
        for groupB_name in groups:

            sample_A = data_subset[data_subset[grouper_column].eq(groupA_name)][
                values_column
            ]
            sample_B = data_subset[data_subset[grouper_column].eq(groupB_name)][
                values_column
            ]
            sample_A = sample_A.astype(np.float64)
            sample_B = sample_B.astype(np.float64)
            assert sample_A.shape[0] == sample_B.shape[0]
            differences = (
                sample_A - sample_B.values
            )  # only the .values of one vector are needed!
            basic_stats = ttest_paired_difference_calculate(differences, conflevel)
            basic_stats.update(
                {
                    "Group A name": groupA_name,
                    "Group B name": groupB_name,
                    "Group A number": sample_A.shape[0],
                    "Group B number": sample_B.shape[0],
                    "Group A average": sample_A.mean(),
                    "Group B average": sample_B.mean(),
                }
            )
            output = output.append(basic_stats, ignore_index=True)

    return output


def confidence_interval(
    df: pd.DataFrame, column_name: str, conflevel=0.95, style="robust"
) -> tuple:
    """
    Calculates the confidence interval, returned as a tuple, for the `column_name` (str) in the
    dataframe `df`, for a given confidence level `conflevel` (default: 0.95).

    `style`: ['robust'; 'regular']: indicates which style of estimates to use for the center and
                                    spread. Default: 'robust'

    Missing values are ignored.
    """

    # TODO : http://www.rips-irsp.com/article/10.5334/irsp.82/

    data = df[column_name]
    n = data.count()

    if style.lower() == "robust":
        center = data.median()
        spread = median_abs_deviation(data.values, nan_policy="omit")
    else:
        center = data.mean()
        spread = data.std()

    c_t = t_value(1 - (1 - conflevel) / 2, n - 1)
    return (center - c_t * spread / np.sqrt(n), center + c_t * spread / np.sqrt(n))


def _mad_1d(x, center, nan_policy):
    """

    Taken from `scipy.stats.stats`

    Median absolute deviation for 1-d array x.
    This is a helper function for `median_abs_deviation`; it assumes its
    arguments have been validated already.  In particular,  x must be a
    1-d numpy array, center must be callable, and if nan_policy is not
    'propagate', it is assumed to be 'omit', because 'raise' is handled
    in `median_abs_deviation`.
    No warning is generated if x is empty or all nan.
    """
    isnan = np.isnan(x)
    if isnan.any():
        if nan_policy == "propagate":
            return np.nan
        x = x[~isnan]
    if x.size == 0:
        # MAD of an empty array is nan.
        return np.nan
    # Edge cases have been handled, so do the basic MAD calculation.
    med = center(x)
    mad = np.median(np.abs(x - med))
    return mad


def median_abs_deviation(
    x, axis=0, center=np.median, scale="normal", nan_policy="omit"
):
    r"""

    Taken from `scipy.stats.stats`: we want the same functionality, but with a slightly different
    default function signature:

    *   `scale='normal'` instead of `scale=1.0`.
    *   `nan_policy='omit'` instead of `nan_policy='propogate'`

    Compute the median absolute deviation of the data along the given axis.  The median absolute
    deviation (MAD_) computes the median over the absolute deviations from the median. It is a
    measure of dispersion similar to the standard deviation but more `robust to outliers`_.
    The MAD of an empty array is ``np.nan``.

    Parameters
    ----------

    x : array_like
        Input array or object that can be converted to an array.
    axis : int or None, optional
        Axis along which the range is computed. Default is 0. If None, compute
        the MAD over the entire array.
    center : callable, optional
        A function that will return the central value. The default is to use
        np.median. Any user defined function used will need to have the
        function signature ``func(arr, axis)``.
    scale : scalar or str, optional
        The numerical value of scale will be divided out of the final
        result. The default is 1.0. The string "normal" is also accepted,
        and results in `scale` being the inverse of the standard normal
        quantile function at 0.75, which is approximately 0.67449.
        Array-like scale is also allowed, as long as it broadcasts correctly
        to the output such that ``out / scale`` is a valid operation. The
        output dimensions depend on the input array, `x`, and the `axis`
        argument.
    nan_policy : {'propagate', 'raise', 'omit'}, optional
        Defines how to handle when input contains nan.
        The following options are available (default is 'propagate'):
        * 'propagate': returns nan
        * 'raise': throws an error
        * 'omit': performs the calculations ignoring nan values

    Returns
    -------
    mad : scalar or ndarray
        If ``axis=None``, a scalar is returned. If the input contains
        integers or floats of smaller precision than ``np.float64``, then the
        output data-type is ``np.float64``. Otherwise, the output data-type is
        the same as that of the input.

    See Also
    --------
    numpy.std, numpy.var, numpy.median, scipy.stats.iqr, scipy.stats.tmean,
    scipy.stats.tstd, scipy.stats.tvar

    Notes
    -----
    The `center` argument only affects the calculation of the central value
    around which the MAD is calculated. That is, passing in ``center=np.mean``
    will calculate the MAD around the mean - it will not calculate the *mean*
    absolute deviation.

    The input array may contain `inf`, but if `center` returns `inf`, the
    corresponding MAD for that data will be `nan`.

    References
    ----------

    .. _MAD:
        "Median absolute deviation",
        https://en.wikipedia.org/wiki/Median_absolute_deviation

    .. _robust to outliers:
        "Robust measures of scale",
        https://en.wikipedia.org/wiki/Robust_measures_of_scale

    Examples
    --------

    When comparing the behavior of `median_abs_deviation` with ``np.std``,
    the latter is affected when we change a single value of an array to have an
    outlier value while the MAD hardly changes:

    >>> from scipy import stats
    >>> x = stats.norm.rvs(size=100, scale=1, random_state=123456)
    >>> x.std()
    0.9973906394005013
    >>> stats.median_abs_deviation(x)
    0.82832610097857
    >>> x[0] = 345.6
    >>> x.std()
    34.42304872314415
    >>> stats.median_abs_deviation(x)
    0.8323442311590675

    Axis handling example:

    >>> x = np.array([[10, 7, 4], [3, 2, 1]])
    >>> x
    array([[10,  7,  4],
           [ 3,  2,  1]])
    >>> stats.median_abs_deviation(x)
    array([3.5, 2.5, 1.5])
    >>> stats.median_abs_deviation(x, axis=None)
    2.0

    Scale normal example:

    >>> x = stats.norm.rvs(size=1000000, scale=2, random_state=123456)
    >>> stats.median_abs_deviation(x)
    1.3487398527041636
    >>> stats.median_abs_deviation(x, scale='normal')
    1.9996446978061115
    """
    if not callable(center):
        raise TypeError(
            "The argument 'center' must be callable. The given "
            f"value {repr(center)} is not callable."
        )

    # An error may be raised here, so fail-fast, before doing lengthy
    # computations, even though `scale` is not used until later
    if isinstance(scale, str):
        if scale.lower() == "normal":
            scale = 0.6744897501960817  # special.ndtri(0.75)
        else:
            raise ValueError(f"{scale} is not a valid scale value.")

    x = np.asarray(x)

    # Consistent with `np.var` and `np.std`.
    if not x.size:
        if axis is None:
            return np.nan
        nan_shape = tuple(item for i, item in enumerate(x.shape) if i != axis)
        if nan_shape == ():
            # Return nan, not array(nan)
            return np.nan
        return np.full(nan_shape, np.nan)  # pragma: no cover

    contains_nan, nan_policy = _contains_nan(x, nan_policy)

    if contains_nan:
        if axis is None:
            mad = _mad_1d(x.ravel(), center, nan_policy)
        else:
            mad = np.apply_along_axis(_mad_1d, axis, x, center, nan_policy)
    else:
        if axis is None:
            med = center(x, axis=None)
            mad = np.median(np.abs(x - med))
        else:
            # Wrap the call to center() in expand_dims() so it acts like
            # keepdims=True was used.
            med = np.expand_dims(center(x, axis=axis), axis)
            mad = np.median(np.abs(x - med), axis=axis)

    return mad / scale


def summary_stats(x, method="robust") -> dict:
    """
    Returns summary statistics of the numeric values in vector `x`.

    Arguments:

        x (Numpy vector or Pandas series): A vector of univariate values to summarize.

    Returns:

        dict: a summary of the univariate vector. The following outputs are the most interesting:

            "center":   a measure of the center  (average). If method is ``robust``, this is the
                        median.
            "spread":   a measure of the spread. If method is ``robust``, this is the Sn, a robust
                        spread estimate.
    """
    if isinstance(x, pd.Series):
        x = x.copy(deep=True).values
    elif isinstance(x, np.ndarray):
        x = x.ravel()
    else:
        raise ValueError("Expecting a NumPy vector or Pandas series.")

    out = {}
    out["mean"] = np.nanmean(x)
    out["std_ddof0"] = np.nanstd(x, ddof=0)
    out["std_ddof1"] = np.nanstd(x, ddof=1)
    out["rsd_classical"] = out["std_ddof1"] / out["mean"]
    (
        out["percentile_05"],
        out["percentile_25"],
        out["median"],
        out["percentile_75"],
        out["percentile_95"],
    ) = np.nanpercentile(x, [5, 25, 50, 75, 95])
    out["iqr"] = out["percentile_75"] - out["percentile_25"]
    out["min"] = np.nanmin(x)
    out["max"] = np.nanmax(x)
    out["N_non_missing"] = np.sum(~np.isnan(x))

    if method.lower() == "robust":
        out["center"], out["spread"] = out["median"], Sn(x)
        if (
            ((out["max"] - out["min"]) > 0)
            and (out["spread"] == 0)
            and (out["N_non_missing"] > 0)
        ):
            # Don't fully trust the Sn() yet. It works strangely on quantized data when there is
            # little variation. Replace the RSD with the classically calculated version in this
            # very specific case. This example shows it: [99, 95, 95, 100, 100, 100, 100, 95, 100,
            # 100, 100, 100, 105, 105, 100, 95, 105, 100, 95, 100]
            out["center"], out["spread"] = out["mean"], out["std_ddof1"]

    else:
        out["center"], out["spread"] = out["mean"], out["std_ddof1"]

    out["rsd"] = out["spread"] / out["center"]
    return out


def outlier_detection_multiple(
    x, algorithm="esd", max_outliers_detected=1, **kwargs
) -> Tuple[List[int], DefaultDict[Any, Any]]:
    """
    Returns a list of indexes of points in the vector `x` which are likely outliers.

    A second output (can be ignored) contains the details of the values used to make the decision.

    Arguments:

        x {list, sequence, NumPy vector/array} -- [A sequence, list or vector which can be
        unravelled.]

    Keyword Arguments:

        algorithm -- Two algorithms are possible to detect outliers: (default: "esd")

            'esd': Generalized ESD Test for Outliers.
            If `max_outliers_detected=1` this is essentially Grubb's test.
            For more details, please see:
            https://www.itl.nist.gov/div898/handbook/eda/section3/eda35h3.htm

            'cc-robust': Build a robust control-chart for the sequence `x` and
            points should lie outside the +/- 3 sigma limits are considered
            outliers.
            Not Implemented Yet: left here as an idea for the future, but not confirmed yet.

        max_outliers_detected -- The maximum number of outliers that
            should be detected, as required by the algorithms.

        kwargs -- Algorithm dependent arguments. Defaults are shown here.

            'esd':

                'robust_variant' = True. Uses the median and MAD for the center
                and the standard deviation respectively.

                'alpha' = 0.05. The significance level of the testing.


    """
    algorithm = algorithm.strip().lower()
    x = pd.Series(x).reset_index(drop=True)
    max_outliers_detected = int(max_outliers_detected)

    if algorithm == "esd":
        """
        https://www.itl.nist.gov/div898/handbook/eda/section3/eda35h3.htm

        Note: the two-sided test is implemented here. p = 1-alpha/(2*(n-i+1))
        """

        # 0. Default settings
        robust_variant = bool(kwargs.get("robust_variant", True))
        alpha = float(kwargs.get("alpha", 0.05))

        assert alpha <= 1.0
        assert max_outliers_detected <= len(x)

        # 1. Run K-S test first to check normality

        # 2. https://www.itl.nist.gov/div898/handbook/eda/section3/eda35h3.htm
        x = x.copy(deep=True)
        extra_out = defaultdict(list)
        N = x.shape[0] - pd.isna(x).sum()

        for k in range(max_outliers_detected):

            i = k + 1
            extra_out["i"].append(i)

            if robust_variant:
                variation = median_abs_deviation(x)
                R = ((x - x.median()) / variation).abs()
            else:
                variation = x.std()
                R = ((x - x.mean()) / variation).abs()

            dof = N - i - 1
            p = 1 - alpha / (2 * (N - i + 1))
            t_s = t_value(p, dof)
            lambda_i = (N - i) * t_s / np.sqrt((dof + np.power(t_s, 2)) * (dof + 2))
            extra_out["lambda"].append(lambda_i)

            R_i_idx = R.idxmax()
            g = R.max()
            # Formula from the R-function for the Grubb's calculation
            t = np.sqrt((g ** 2 * N * (2 - N)) / (g ** 2 * N - (N - 1) ** 2))
            if t <= 0:
                p_value = 0
            else:
                p_value = min(N * (1 - t_value_cdf(t, N - 2)), 1)

            extra_out["R_i_idx"].append(R_i_idx)
            extra_out["R_i"].append(R.max())
            extra_out["p-value"].append(p_value)

            if variation > __eps:
                # The variation, if zero or small, will fail to drop this index
                x.drop(R_i_idx, inplace=True)

        try:
            cutoff_i = np.where(
                np.array(extra_out["R_i"]) - np.array(extra_out["lambda"]) >= 0,
            )[0][0]
        except IndexError:
            cutoff_i = -1

        # The outlier indices are the points from the start of
        # `extra_out['R_i_idx']`, up to, and including, the `cutoff_i` index
        # in the list.
        extra_out["cutoff"] = cutoff_i

        outlier_index = extra_out["R_i_idx"][0 : (cutoff_i + 1)]
        return outlier_index, extra_out
    else:
        return [], defaultdict(dict)


def within_between_standard_deviation(df, measured: str, repeat: str) -> dict:
    """
    Given a DataFrame `df` of raw data, and an indication of which column is the `measured` value
    column, and which is the `repeat` indicator, it will calculate the within and between replicate
    standard deviation.

    Example

    Two measurements on day 1 ``[101, 102]`` and two measurements on day 2 ``[94, 95]``. The
    between-day variation can already be expected to be much greater than the within-day variation.

    >>> df = pd.DataFrame(data={'Result': [101, 102, 94, 95], 'Repeat': [1, 1, 2, 2]})
        Result  Repeat
    0     101       1
    1     102       1
    2      94       2
    3      95       2
    >>> output = within_between_standard_deviation(df, measured="Result", repeat="Repeat")
    {'total_ms':       16.666667,
     'total_dof':      3,
     'within_ms':      0.5,
     'within_stddev':  0.70711,
     'within_dof':     2,
     'between_ms':     49.0,
     'between_stddev': 7.0,
     'between_dof':    1}

    Note

    * SSQ = sum of squares
    * DOF= degrees of freedom
    * MS = mean square = (sum of squares) / (degrees of freedom) = SSQ / DOF = variance
    """

    # Overall statistics:
    total_ms = max(0, df[measured].var())
    total_dof = max(0, df[measured].count() - 1)

    # Within a group: calculate the standard deviation (of variance) of each group. If you take
    # the average of those variances, pooled, you get an estimate of the within-group variance.
    within_ms = 0.0
    within_dof = 0
    for idx, group in df.groupby(repeat):
        dof_group_i = max(0, group[measured].count() - 1)
        within_dof += dof_group_i
        # handling missing data makes for messier code
        within_ms = np.nansum([within_ms, group[measured].var() * dof_group_i])

    if within_dof == 0:
        within_ms = 0
    else:
        within_ms /= within_dof

    # Between groups: the ANOVA relationship is such that SSQ(total) = SSQ(between) + SSQ(within).
    # Therefore the between-group statistics are found by differencing
    between_dof = max(0, total_dof - within_dof)
    if between_dof == 0:
        between_ms = 0
    else:
        between_ms = max(
            0.0, (total_ms * total_dof - within_ms * within_dof) / between_dof
        )

    return {
        "total_ms": total_ms,
        "total_dof": total_dof,
        "within_ms": within_ms,
        "within_stddev": np.sqrt(within_ms),
        "within_dof": within_dof,
        "between_ms": between_ms,
        "between_stddev": np.sqrt(between_ms),
        "between_dof": between_dof,
    }



# Built-in libraries
import json
import math
import random
from typing import Any, Dict

# Plotting settings
import plotly.graph_objects as go
from plotly.offline import plot as plotoffline
import seaborn as sns

import numpy as np

# import pandas as pd

from .data_input import check_valid_batch_dict


def get_rgba_from_triplet(incolour, alpha=1, as_string=False):
    """
    Convert the input colour triplet (list) to a Plotly rgba(r,g,b,a) string if
    `as_string` is True. If `False` it will return the list of 3 integer RGB
    values.

    E.g.    [0.9677975592919913, 0.44127456009157356, 0.5358103155058701] -> 'rgba(246,112,136,1)'
    """
    assert len(incolour) == 3
    colours = [max(0, int(math.floor(c * 255))) for c in incolour]
    if as_string:
        return f"rgba({colours[0]},{colours[1]},{colours[2]},{float(alpha)})"
    else:
        return colours


def plot_to_HTML(filename: str, fig: dict):
    config = dict(
        {
            "scrollZoom": True,
            "displayModeBar": True,
            "editable": False,
            "displaylogo": False,
            "showLink": False,
        }
    )
    return plotoffline(
        figure_or_data=fig,
        config=config,
        filename=filename,
        include_plotlyjs="cdn",
        include_mathjax="cdn",
        auto_open=False,
    )


def plot__all_batches_per_tag(
    df_dict: dict,
    tag: str,
    tag_y2: str = None,
    time_column: str = None,
    extra_info="",
    batches_to_highlight={},
    x_axis_label: str = "Time [sequence order]",
    highlight_width: int = 5,
    html_image_height: int = 900,
    html_aspect_ratio_w_over_h: float = 16 / 9,
    y1_limits: tuple = (None, None),
    y2_limits: tuple = (None, None),
) -> go.Figure:
    """Plots a particular `tag` over all batches in the given dataframe `df`.

    Parameters
    ----------
    df_dict : dict
        Standard data format for batches.
    tag : str
        Which tag to plot? [on the y1 (left) axis]
    tag_y2 : str, optional
        Which tag to plot? [on the y2 (right) axis]
        Tag will be plotted with different scaling on the secondary axis, to allow time-series
        comparisons to be easier.
    time_column : str, optional
        Which tag on the x-axis. If not specified, creates sequential integers, starting from 0
        if left as the default, `None`.
    extra_info : str, optional
        Used in the plot title to add any extra details, by default ""
    batches_to_highlight : dict, optional
        keys: an string which can be json.loads(...) and turns into a Plotly line specifier.
        For example:
            batches_to_highlight = grouper= {'{"width": 2, "color": "rgba(255,0,0,0.5)"}': redlist}

            will plot batch identifiers (must be valid keys in `df_dict`) in the "redlist" list
            with that colour and linewidth.
    x_axis_label : str, optional
        String label for the x-axis, by default "Time [sequence order]"
    highlight_width: int, optional
        The width of the highlighted lines; default = 5.
    html_image_height : int, optional
        HTML image output height, by default 900
    html_aspect_ratio_w_over_h : float, optional
        HTML image aspect ratio: 16/9 (therefore the default width will be 1600 px)
    y1_limits: tuple, optional
        Axis limits enforced on the y1 (left) axis. Default is (None, None) which means the data
        themselves are used to determine the limits. Specify BOTH limits. Plotly requires
        (at the moment https://github.com/plotly/plotly.js/issues/400) that you specify both.
        Order: (low limit, high limit)
    y2_limits: tuple, optional
        Axis limits enforced on the y2 (right) axis. Default is (None, None) which means the data
        themselves are used to determine the limits. Specify BOTH limits. Plotly requires
        (at the moment https://github.com/plotly/plotly.js/issues/400) that you specify both.


    Returns
    -------
    go.Figure
        Standard Plotly fig object (dictionary-like).
    """
    default_line_width = 2
    unique_items = list(df_dict.keys())
    n_colours = len(unique_items)
    random.seed(13)
    colours = list(sns.husl_palette(n_colours))
    random.shuffle(colours)
    colours = [get_rgba_from_triplet(c, as_string=True) for c in colours]
    line_styles = {
        k: dict(width=default_line_width, color=v)
        for k, v in zip(unique_items, colours)
    }
    for key, val in batches_to_highlight.items():
        line_styles.update(
            {item: json.loads(key) for item in val if item in df_dict.keys()}
        )

    highlight_list = []
    for key, val in batches_to_highlight.items():
        highlight_list.extend(val)

    highlight_list = list(set(highlight_list))

    fig = go.Figure()

    for batch_id, batch_df in df_dict.items():
        assert (
            tag in batch_df.columns
        ), f"Tag '{tag}' not found in the batch with id {batch_id}."
        if tag_y2:
            assert (
                tag_y2 in batch_df.columns
            ), f"Tag '{tag}' not found in the batch with id {batch_id}."
        if time_column in batch_df.columns:
            time_data = batch_df[time_column]
        else:
            time_data = list(range(batch_df.shape[0]))

        if batch_id in highlight_list:
            continue  # come to this later
        else:
            fig.add_trace(
                go.Scatter(
                    x=time_data,
                    y=batch_df[tag],
                    name=batch_id,
                    line=line_styles[batch_id],
                    mode="lines",
                    opacity=0.8,
                    yaxis="y1",
                )
            )
            if tag_y2:
                fig.add_trace(
                    go.Scatter(
                        x=time_data,
                        y=batch_df[tag_y2],
                        name=batch_id,
                        line=line_styles[batch_id],
                        mode="lines",
                        opacity=0.8,
                        yaxis="y2",
                    )
                )

    # Add the highlighted batches last: therefore, sadly, we have to do another run-through.
    # Plotly does not yet support z-orders.
    for batch_id, batch_df in df_dict.items():
        if time_column in batch_df.columns:
            time_data = batch_df[time_column]
        else:
            time_data = list(range(batch_df.shape[0]))

        if batch_id in highlight_list:
            fig.add_trace(
                go.Scatter(
                    x=time_data,
                    y=batch_df[tag],
                    line=line_styles[batch_id],
                    name=batch_id,
                    mode="lines",
                    opacity=0.8,
                    yaxis="y1",
                )
            )
            if tag_y2:
                fig.add_trace(
                    go.Scatter(
                        x=time_data,
                        y=batch_df[tag_y2],
                        line=line_styles[batch_id],
                        name=batch_id,
                        mode="lines",
                        opacity=0.8,
                        yaxis="y2",
                    )
                )

    yaxis1_dict = dict(
        title=tag, gridwidth=2, matches="y1", showticklabels=True, side="left"
    )
    if (y1_limits[0] is not None) or (y1_limits[1] is not None):
        yaxis1_dict["autorange"] = False
        yaxis1_dict["range"] = y1_limits

    yaxis2_dict: Dict[str, Any] = dict(
        title=tag_y2, gridwidth=2, matches="y2", showticklabels=True, side="right"
    )
    if (y2_limits[0] is not None) or (y2_limits[1] is not None):
        yaxis2_dict["autorange"] = False
        yaxis2_dict["range"] = y2_limits

    fig.update_layout(
        title=f"Plot of: '{tag}'"
        + (f" on left axis; with '{str(tag_y2)}' on right axis." if tag_y2 else ".")
        + (f" [{str(extra_info)}]" if extra_info else ""),
        hovermode="closest",
        showlegend=True,
        legend=dict(
            orientation="h",
            traceorder="normal",
            font=dict(family="sans-serif", size=12, color="#000"),
            bordercolor="#DDDDDD",
            borderwidth=1,
        ),
        autosize=False,
        xaxis=dict(title=x_axis_label, gridwidth=1),
        yaxis=yaxis1_dict,
        width=html_aspect_ratio_w_over_h * html_image_height,
        height=html_image_height,
    )
    if tag_y2:
        fig.update_layout(yaxis2=yaxis2_dict)
    return fig


def plot__multitags(
    df_dict: dict,
    batch_list: list = None,
    tag_list: list = None,
    time_column: str = None,
    batches_to_highlight: dict = {},
    settings: dict = None,
    fig=None,
) -> go.Figure:
    """
    Plots all the tags for a batch; or a subset of tags, if specified in `tag_list`.

    Parameters
    ----------
    df_dict : dict
        Standard data format for batches.

    batch_list : list [default: None, will plot all batches in df_dict]
        Which batches to plot; if provided, must be a list of valid keys into df_dict.

    tag_list : list [default: None, will plot all tags in the dataframes]
        Which tags to plot; tags will also be plotted in this order, or in the order of the
        first dataframe if not specified.

    time_column : str, optional
        Which tag on the x-axis. If not specified, creates sequential integers, starting from 0
        if left as the default, `None`.

    batches_to_highlight : dict, optional
        keys: an string which can be json.loads(...) and turns into a Plotly line specifier.
        For example:
            batches_to_highlight = grouper= {'{"width": 2, "color": "rgba(255,0,0,0.5)"}': redlist}

            will plot batch identifiers (must be valid keys in `df_dict`) in the "redlist" list
            with that colour and linewidth.

    settings : dict
        Default settings are = {
            "nrows": 1 [int],
                Number of rows in the plot.

            "ncols": None
                None = use as many columns as required to plot the data; else, supply an integer.

            "x_axis_label": "Time, grouped per tag"
                What label is added to the x-axis?

            "title": ""
                Overall plot title

            "showlegend": True,
                Add a legend item for each tag

            "html_image_height": 900,
                in pixels

            "html_aspect_ratio_w_over_h": 16/9,
                sets the image width, as a ratio of the height

        }

    fig : go.Figure
        If supplied, uses the existing Plotly figure to draw in.

    """
    # This will be clumsy, until we have Python 3.9
    default_settings: Dict[str, Any] = dict(
        nrows=1,
        ncols=0,
        x_axis_label="Time, grouped per tag",
        title="",
        showlegend=True,
        html_image_height=900,
        html_aspect_ratio_w_over_h=16 / 9,
        default_line_width=2,
    )
    if settings:
        default_settings.update(settings)

    settings = default_settings

    assert check_valid_batch_dict(df_dict, no_nan=False)

    if fig is None:
        fig = go.Figure()

    batch1 = df_dict[list(df_dict.keys())[0]]
    if tag_list is None:
        tag_list = list(batch1.columns)
    tag_list = list(tag_list)  # Force it; sometimes we get non-list inputs

    if batch_list is None:
        batch_list = list(df_dict.keys())
    batch_list = list(batch_list)

    if time_column in tag_list:
        tag_list.remove(time_column)

    if settings["ncols"] == 0:
        settings["ncols"] = int(np.ceil(len(tag_list) / int(settings["nrows"])))

    specs = [[{"type": "scatter"}] * int(settings["ncols"])] * int(settings["nrows"])

    fig.set_subplots(
        rows=settings["nrows"],
        cols=settings["ncols"],
        shared_xaxes="all",
        shared_yaxes=False,
        start_cell="top-left",
        vertical_spacing=0.2 / settings["nrows"],
        horizontal_spacing=0.2 / settings["ncols"],
        subplot_titles=tag_list,
        specs=specs,
    )

    n_colours = len(df_dict)
    random.seed(13)
    colours = list(sns.husl_palette(n_colours))
    colours = [get_rgba_from_triplet(c, as_string=True) for c in colours]
    colour_assignment = {
        key: dict(width=settings["default_line_width"], color=val)
        for key, val in zip(list(df_dict.keys()), colours)
    }
    for key, val in batches_to_highlight.items():
        colour_assignment.update(
            {item: json.loads(key) for item in val if item in df_dict.keys()}
        )
    margin_dict = dict(l=10, r=10, b=5, t=80)  # Defaults: l=80, r=80, t=100, b=80

    for batch_id, batch_df in df_dict.items():
        if batch_id not in batch_list:
            continue
        # Time axis values
        if time_column in batch_df.columns:
            time_data = batch_df[time_column]
        else:
            time_data = list(range(batch_df.shape[0]))

        row = col = 1
        for tag in tag_list:
            trace = go.Scatter(
                x=time_data,
                y=batch_df[tag],
                name=batch_id,
                mode="lines",
                hovertemplate="Time: %{x}\ny: %{y}",
                line=colour_assignment[batch_id],
                legendgroup=batch_id,
                showlegend=settings["showlegend"] if tag == tag_list[0] else False,
            )
            fig.add_trace(trace, row=row, col=col)

            col += 1
            if col > settings["ncols"]:
                row += 1
                col = 1

    fig.update_layout(
        title=settings["title"],
        margin=margin_dict,
        hovermode="closest",
        showlegend=settings["showlegend"],
        legend=dict(
            orientation="h",
            traceorder="normal",
            font=dict(family="sans-serif", size=12, color="#000"),
            bordercolor="#DDDDDD",
            borderwidth=1,
        ),
        autosize=False,
        xaxis=dict(
            title=settings["x_axis_label"],
            gridwidth=1,
            mirror=True,  # ticks are mirror at the top of the frame also
            showspikes=True,
            visible=True,
        ),
        yaxis=dict(
            gridwidth=2,
            type="linear",
            autorange=True,
            showspikes=True,
            visible=True,
            showline=True,  # show a separating line
            side="left",  # show on the RHS
        ),
        width=settings["html_aspect_ratio_w_over_h"] * settings["html_image_height"],
        height=settings["html_image_height"],
    )

    return fig



# (c) Kevin Dunn, 2019-2021. MIT License.

"""
Various factorial designs
"""
from typing import Optional

from .structures import create_names, c, expand_grid


def full_factorial(nfactors: int, names: Optional[list] = None):
    """
    Creates a full factorial (2^k) design for the case when there are
    `nfactors` [integer] number of factors.

    The optional list of `names` can be provided. The entries in the list
    should be strings. If not provided, the names will be created.
    """
    nfactors = int(nfactors)
    if names is None:
        names = create_names(nfactors)

    # Expand the full factorial out into variables
    return expand_grid(**dict(zip(names, [c(-1, +1)] * len(names))))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="process-improve",
    version="0.9.16",
    author="Kevin Dunn",
    author_email="kgdunn@gmail.com",
    description="Process Improvement using Data: Designed Experiments; Latent Variables (PCA, PLS, multivariate methods with missing data); Process Monitoring; Batch data analysis.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kgdunn/process_improve",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    install_requires=[
        "pandas",
        "numpy",
        "statsmodels",
        "matplotlib",
        "bokeh",
        "sklearn",
        "patsy",
        "scikit-image",
    ],
    # Include additional files into the package
    include_package_data=True,
)



import numpy as np
import pandas as pd
import statsmodels.api as sm
from ..univariate.metrics import t_value

__eps = np.finfo(np.float32).eps


def fit_robust_lm(x: np.ndarray, y: np.ndarray) -> list:
    """
    Fits a robust linear model between Numpy vectors `x` and `y`, with an
    intercept. Returns a list: [intercept, slope] of the fit. No extra
    checking on data consistency is done.

    See also:

    regression.repeated_median_slope
    """
    rlm_model = sm.RLM(
        y, np.vstack([np.ones(x.size), x.ravel()]).T, M=sm.robust.norms.HuberT()
    )
    rlm_results = rlm_model.fit()
    return rlm_results.params


def repeated_median_slope(x, y, nowarn=False):
    """
    Robust slope calculation.

    https://en.wikipedia.org/wiki/Repeated_median_regression

    An elegant (simple) method to compute the robust slope between a vector `x` and `y`.

    INVESTIGATE: algorithm speed-ups via these articles:
    https://link.springer.com/article/10.1007/PL00009190
    http://www.sciencedirect.com/science/article/pii/S0020019003003508
    """
    # Slope
    medians = []
    x = x.copy().ravel()
    y = y.copy().ravel()
    if not (nowarn):
        assert len(x) > 2, "More than two samples are required for this function."
        assert len(x) == len(y), "Vectors x and y must have the same length."

    for i in np.arange(len(x)):
        inner_medians = []
        for j in np.arange(len(y)):
            den = x[j] - x[i]
            if j != i and den != 0:
                inner_medians.append((y[j] - y[i]) / den)

        medians.append(np.nanmedian(inner_medians))

    return np.nanmedian(medians)


def simple_robust_regression(
    x, y, na_rm=None, conflevel=0.95, nowarn=False, pi_resolution=50
):
    """
    x and y: iterables
    na_rm: None; no effect for robust regression. Here for consistency with the non-robust case.
    nowarn: If True, then no error checking/warnings are issued. The user is committing to do that
    themselves ahead of time.

    Simple robust regression between an `x` and a `y` using the `repeated_median_slope` method
    to calculate the slope. The intercept is the median intercept, when using that slope and the
    provided `x` and `y` values.

    The rest of the classical output from a regression are based on these robust parameter
    estimates.
    """
    x_, y_ = x.copy().ravel(), y.copy().ravel()
    x, y = x_[~np.isnan(x_) & ~np.isnan(y_)], y_[~np.isnan(x_) & ~np.isnan(y_)]

    slope = repeated_median_slope(x, y, nowarn=nowarn)
    intercept = np.nanmedian(y - slope * x)
    mean_x, mean_y = np.mean(x), np.mean(y)
    total_ssq = np.sum(np.power(y - mean_y, 2))

    out = {}

    out["N"] = min(
        x.size - np.count_nonzero(np.isnan(x)), y.size - np.count_nonzero(np.isnan(y))
    )
    out["intercept"] = intercept
    out["coefficients"] = [
        slope,
    ]
    out["fitted_values"] = intercept + slope * x
    regression_ssq = np.sum(np.power(out["fitted_values"] - mean_y, 2))
    out["residuals"] = y - out["fitted_values"]
    residual_ssq = np.sum(out["residuals"] * out["residuals"])
    # For robust method, calculate this way, since no guarantee both RegSS or RSS are < TSS.
    # So ensure this way that the TSS = RegSS + RSS, and R2 is the ratio of RegSS/TSS
    # https://learnche.org/pid/least-squares-modelling/least-squares-model-analysis
    total_ssq = regression_ssq + residual_ssq
    out["R2"] = regression_ssq / total_ssq
    out["SE"] = np.sqrt(residual_ssq / (len(x) - 2))
    out["x_ssq"] = np.sum(np.power(x - np.mean(x), 2))
    c_t = t_value(1 - (1 - conflevel) / 2, out["N"] - 2)  # 2 fitted parameters
    # out["t_value"] = np.array([c_t])  # for consistency with other regression models.
    # "pi" = prediction interval
    pi_range = np.linspace(np.min(x), np.max(x), pi_resolution)
    pi_y_pred = out["intercept"] + out["coefficients"][0] * pi_range
    if out["x_ssq"] < __eps:
        out["standard_error_intercept"] = SE_b0 = np.NaN
        out["standard_errors"] = [
            np.NaN,
        ]
        out["pi_range"] = np.vstack([pi_range, pi_y_pred, pi_y_pred]).T

    else:
        out["standard_error_intercept"] = SE_b0 = out["SE"] * np.sqrt(
            (1 / out["N"] + (mean_x) ** 2 / out["x_ssq"])
        )
        out["standard_errors"] = [
            out["SE"] * 1 / np.sqrt(out["x_ssq"]),
        ]
        var_y = (out["SE"] ** 2) * (
            1 + 1 / out["N"] + (pi_range - np.mean(x)) ** 2 / out["x_ssq"]
        )
        std_y = np.sqrt(var_y)
        lower = pi_y_pred - c_t * std_y
        upper = pi_y_pred + c_t * std_y
        out["pi_range"] = np.vstack([pi_range, lower, upper]).T

    out["conf_interval_intercept"] = np.array(
        [out["intercept"] - c_t * SE_b0, out["intercept"] + c_t * SE_b0],
    )
    out["conf_intervals"] = np.array(
        [
            [
                out["coefficients"][0] - c_t * out["standard_errors"][0],
                out["coefficients"][0] + c_t * out["standard_errors"][0],
            ],
        ]
    )
    out["leverage"] = 1 / out["N"] + np.power(x - mean_x, 2) / out["x_ssq"]
    out["k"] = 2
    if out["SE"] < __eps:
        out["influence"] = out["residuals"] * 0.0
    else:
        out["influence"] = (
            np.power(out["residuals"] / ((1 - out["leverage"]) * out["SE"]), 2)
            * out["leverage"]
            / out["k"]
        )

    return out


def multiple_linear_regression(
    X, y, fit_intercept=True, na_rm=True, conflevel=0.95, pi_resolution=50
):
    """
    Linear regression of the N rows and K columns of matrix `X` onto the single column 'y'.

    The matrix `X` will be augmented with a column of 1's if `fit_intercept` is True.
    `na_rm`: True:  removes all observations with one or more missing values.

    Notes and limitations:
        * does not handle weighting
        * N >= K at least as many rows as columns in X

    Returns a dictionary of outputs with these keys:

        coefficients:   a vector of K coefficients, one for each column in ``X``

        intercept:      returned if ``fit_intercept==True``

        standard_errors:a vector of K standard errors, one for each column in ``X``

        standard_error_intercept: standard error for the intercept

        R2:             the infamous R^2 values

        SE              the model's standard error

        fitted_values   the N predicted values, one per row in ``y``

        residuals       the N residuals

        t_value         the t-values for the standard errors

        conf_intervals  the 95% confidence intervals for the model terms: K rows,
                        2 columns: column 1 is lower, column 2 is upper

        pi_range       the prediction intervals, above an below, over the range of data.

    TODO: report hatvalues, discrepancy:  for residual detection
    """

    alpha = 1.0 - conflevel
    out = {
        "N": None,
        "coefficients": [
            np.nan,
        ],
        "intercept": np.nan,
        "standard_errors": [
            np.nan,
        ],
        "standard_error_intercept": np.nan,
        "R2": np.nan,
        "SE": np.nan,
        "fitted_values": np.nan,
        "residuals": np.nan,
        "t_value": np.nan,
        "conf_intervals": [np.nan, np.nan],
    }

    #  Data pre-processing: handle both Pandas and NumPy arrays
    if isinstance(X, np.ndarray):
        X = pd.DataFrame(X, copy=True)
    else:
        X = pd.DataFrame(X.values, copy=True)

    if isinstance(y, np.ndarray):
        y = pd.DataFrame(y.ravel(), copy=True)
    else:
        y = pd.DataFrame(y.values, copy=True)

    # Removing missing values:
    if na_rm:
        missing_idx = y.isna().any(axis=1) | X.isna().any(axis=1)
        X = X.loc[~missing_idx, :]
        y = y.loc[~missing_idx]

    # CASE when there is no data, or only 1 data point
    if (y.size <= 1) or (X.size <= 1):
        return out

    #  Hard checks: to be converted to graceful errors later on
    out["N"], k = X.shape
    mean_y = np.mean(y.values)
    total_ssq = np.sum(np.power(y.values - mean_y, 2))
    x_vector = X.copy()
    if fit_intercept:
        X = sm.add_constant(X)
        k = k + 1

    assert (
        out["N"] >= k
    ), "N >= K: You need at least as many rows as there are columns to fit a linear regression."
    assert out["N"] == y.size

    if x_vector.shape[1] == 1:
        mean_X = np.mean(x_vector)
        out["x_ssq"] = np.sum(np.power(x_vector - mean_X, 2))[0]

        # Can be calculated before the model is even fit:
        out["leverage"] = (
            1 / out["N"] + np.power(x_vector - mean_X, 2) / out["x_ssq"]
        ).values.ravel()

    # Do the work
    model = sm.OLS(y, X)
    results = model.fit()

    #  Report the results:
    if fit_intercept:
        out["intercept"] = results.params.values[0]
        out["standard_error_intercept"] = results._results.bse[0]
        out["coefficients"] = results.params.values[1:]
        out["standard_errors"] = results._results.bse[1:]
        out["t_value"] = results.tvalues[1:]
        out["conf_intervals"] = results._results.conf_int(alpha)[1:, :]
        out["conf_interval_intercept"] = results._results.conf_int(alpha)[0, :]
    else:
        out["coefficients"] = results.params[:]
        out["t_value"] = results.tvalues[:]
        out["standard_errors"] = results._results.bse[:]
        out["conf_intervals"] = results._results.conf_int(alpha)

    out["fitted_values"] = results._results.fittedvalues
    out["R2"] = results._results.rsquared
    regression_ssq = np.sum(np.power(out["fitted_values"] - mean_y, 2))
    out[
        "residuals"
    ] = results._results.resid  # == y.values.ravel() - out["fitted_values"]
    residual_ssq = np.sum(out["residuals"] * out["residuals"])
    out["R2_regression_based"] = regression_ssq / total_ssq
    out["R2_residual_based"] = 1 - (residual_ssq / total_ssq)
    out["SE"] = np.sqrt(results._results.scale)  # np.sqrt(residual_ssq / (len(x) - 2))
    out["k"] = k

    if x_vector.shape[1] == 1 and fit_intercept:
        # "pi" = prediction interval
        pi_range = np.linspace(
            np.min(X.values[:, 1]), np.max(X.values[:, 1]), pi_resolution
        )
        pi_y_pred = out["intercept"] + out["coefficients"][0] * pi_range
        if out["SE"] < __eps:
            out["influence"] = out["residuals"] * 0.0
        else:
            out["influence"] = (
                np.power(out["residuals"] / ((1 - out["leverage"]) * out["SE"]), 2)
                * out["leverage"]
                / k
            )
            var_y = (out["SE"] ** 2) * (
                1
                + 1 / out["N"]
                + (pi_range - np.mean(X.values[:, 1])) ** 2 / out["x_ssq"]
            )
            std_y = np.sqrt(var_y)
            c_t = t_value(1 - (1 - conflevel) / 2, out["N"] - 2)  # 2 fitted parameters
            lower = pi_y_pred - c_t * std_y
            upper = pi_y_pred + c_t * std_y
            out["pi_range"] = np.vstack([pi_range, lower, upper]).T

    return out



# (c) Kevin Dunn, 2010-2021. MIT License. Based on own private work over the years.

import itertools
from collections import defaultdict
from collections.abc import Iterable
import numpy as np
import pandas as pd


class Column(pd.Series):
    """
    Creates a column. Can be used as a factor, or a response vector.
    """

    # https://pandas.pydata.org/pandas-docs/stable/development/extending.html
    # Temporary properties
    _internal_names = pd.DataFrame._internal_names + ["not_used_for_now"]
    _internal_names_set = set(_internal_names)

    # Properties which survive subsetting, etc
    _metadata = [
        "pi_index",  # might be used later if the user provides their own index
        "pi_numeric",  # if numeric indicator
        "pi_lo",  # if numeric: low level (-1)
        "pi_hi",  # if numeric: high level (+1)
        "pi_range",  # if numeric: range: distance from low to high
        "pi_center",  # if numeric: midway between low and high (0)
        "pi_is_coded",  # is it a coded variables, or in real-world units
        "pi_units",  # string variable, containing the units
        "pi_name",  # name of the column
    ]

    @property
    def _constructor(self):
        return Column

    def to_coded(self, center=None, range=None):
        """
        Converts the column vector to coded units.
        """
        out = self.copy(deep=True)
        if self.pi_is_coded:
            return out

        x_center = center or self.pi_center
        x_range = range or self.pi_range

        # Simply override the values and the `pi_is_coded` flag, but all the
        # rest remains as is.
        out.iloc[:] = (self.values - x_center) / (0.5 * np.diff(x_range)[0])
        out.pi_is_coded = True
        if out.pi_name:
            out.name = f"{out.pi_name} [coded]"

        return out

    def to_realworld(self, center=None, range=None):
        """
        Converts the column vector to real-world units.
        """
        out = self.copy(deep=True)
        if not self.pi_is_coded:
            return out

        x_center = center or self.pi_center
        x_range = range or self.pi_range
        # Simply override the values and the `pi_is_coded` flag, but all the
        # rest remains as is.
        out.iloc[:] = self.values * (0.5 * np.diff(x_range)[0]) + x_center
        out.pi_is_coded = False
        if out.pi_name and out.pi_units:
            out.name = f"{out.pi_name} [{out.pi_units}]"

        return out

    def copy(self, deep=True):
        out = pd.Series.copy(self, deep=deep)
        out.name = self.name
        return out

    def extend(self, values):
        """
        Extends the column with the list of new values.
        """
        assert isinstance(values, list), "The 'values' must be in a list [...]"
        prior_n = self.index[-1]
        index = list(range(prior_n + 1, prior_n + len(values) + 1))
        new = pd.Series(data=values, index=index)
        intermediate = self.copy(deep=True)
        intermediate = intermediate.append(new)

        # Carry the meta data over. For some reason the `pd.Series.append`
        # function does not do this (yet?)
        for key in self._metadata:
            setattr(intermediate, key, getattr(self, key))
        intermediate.name = self.name
        return intermediate


class Expt(pd.DataFrame):
    """
    Dataframe object with experimental data. Builds on the Pandas dataframe,
    but with some extra attributes.
    """

    # Temporary properties
    _internal_names = pd.DataFrame._internal_names + ["not_used_for_now"]
    _internal_names_set = set(_internal_names)

    # Properties which survive subsetting, etc
    _metadata = ["pi_source", "pi_title", "pi_units"]

    @property
    def _constructor(self):
        return Expt

    def __repr__(self):
        title = f"Name: {self.pi_title}"
        dimensions = f"Size: {self.shape[0]} experiments; " f"{self.shape[1]} columns."
        return "\n".join([pd.DataFrame.__repr__(self), title, dimensions])

    def get_title(self):
        return self.pi_title or ""


def create_names(n: int, letters=True, prefix="X", start_at=1, padded=True):
    """
    Returns default factor names, for a given number of `n` [integer] factors.
    The factor name "I" is never used.

    If `letters` is True (default), then at most 25 factors can be returned.

    If `letters` is False, then the prefix is used to construct names which are
    the combination of the prefix and numbers, starting at `start_at`.

    Example:
        >>> create_names(5)
            ["A", "B", "C", "D", "E"]

        >>> create_names(3, letters=False)
            ["X1", "X2", "X3"]

        >>> create_names(3, letters=False, prefix='Q', start_at=9,
                             padded=True)
            ["Q09", "Q10", "Q11"]
    """
    if letters and n <= 25:
        out = [chr(65 + i) for i in range(n)]
        if "I" in out:
            out.remove("I")
            out.append(chr(65 + n))

    else:
        longest = 0
        if padded:
            longest = len(str(start_at + n - 1))

        out = [
            f'{str(prefix)}{str(i).rjust(longest, "0")}'
            for i in range(start_at, n + start_at)
        ]

    return out


def c(*args, **kwargs) -> Column:  # noqa: C901
    """
    Performs the equivalent of the R function "c(...)", to combine data elements
    into a DataFrame. Converts every entry into a floating point object.

    Inputs
    ------

    index: a list of names for the entries in `args`

    name:  a name for the column

    Usage
    -----
    # All equivalent ways of creating a factor, "A"

    A = c(-1, 0, +1, -1, +1)

    A = c(-1, 0, +1, -1, +1, index=['lo', 'cp', 'hi', 'lo', 'hi'])
    A = c( 4, 5,  6,  4,  6, range=(4, 6))
    A = c( 4, 5,  6,  4,  6, center=5, range=(4, 6))  # more explicit
    A = c( 4, 5,  6,  4,  6, lo=4, hi=6)
    A = c( 4, 5,  6,  4,  6, lo=4, hi=6, name = 'A')
    A = c([4, 5,  6,  4,  6], lo=4, hi=6, name = 'A')
    A = c([4, 5,  6,  4,  6], lo=4, hi=6, name = 'A')

    # By default, the assumption is the variable levels supplied are coded
    # units. But if any one of the following: `lo`, `hi`, `center`, `range` OR
    # `units` are specified, then immediately it is assumed that the variable
    # values are not coded.
    # So, to force the specification, you may supply the optional input of
    # `coded` as True or False
    A = c([4, 5,  6,  4,  6], lo=1, hi=3, coded=True)
    A = c([4, 5,  6,  4,  6], lo=1, hi=3, coded=False, units="g/mL")

    # Categorical variables
    B = c(0, 1, 0, 1, 0, 2, levels =(0, 1, 2))
    M = c("Dry", "Wet", "Dry", "Wet", levels = ("Dry", "Wet"))

    """
    sanitize = []
    numeric = True
    override_coded = kwargs.get("coded", None)

    if "levels" in kwargs:
        numeric = False

    for j in args:
        if isinstance(j, Iterable):
            if isinstance(j, np.ndarray):
                sanitize = j.ravel().tolist()

            if isinstance(j, pd.Series):
                sanitize = j.copy()
                if "index" not in kwargs:
                    kwargs["index"] = sanitize.index

            if isinstance(j, list):
                sanitize = j.copy()

            try:
                sanitize = [float(j) for j in sanitize]
            except ValueError:
                numeric = False

        else:
            try:
                sanitize.append(float(j))
            except ValueError:
                numeric = False
                sanitize.append(j)

    # Index creation
    default_idx = list(range(1, len(sanitize) + 1))
    index = kwargs.get("index", default_idx)
    if len(index) != len(sanitize):
        raise IndexError(
            ('Length of "index" must match the ' "number of numeric inputs.")
        )

    out = Column(data=sanitize, index=index, name=None)
    # Use sensible defaults, if not provided
    out.pi_index = True
    out.pi_lo = None
    out.pi_hi = None
    out.pi_range = None
    out.pi_center = None
    out.pi_numeric = numeric
    out.pi_units = None
    out.pi_name = None
    out.pi_is_coded = True

    out.pi_name = kwargs.get("name", "Unnamed")
    out.name = out.pi_name
    if numeric:

        # If any of 'lo', 'hi', 'center', or 'range' are specified, then it
        # is assumed that the variable is NOT coded
        try:
            out.pi_lo = kwargs["lo"]
            out.pi_is_coded = False
        except KeyError:
            out.pi_lo = out.min()

        try:
            out.pi_hi = kwargs["hi"]
            out.pi_is_coded = False
        except KeyError:
            out.pi_hi = out.max()

        try:
            out.pi_range = kwargs["range"]
            out.pi_is_coded = False
        except KeyError:
            out.pi_range = (out.pi_lo, out.pi_hi)

        try:
            _ = (e for e in out.pi_range)
        except TypeError:
            assert False, "The `range` input must be an iterable, with " "2 values."
        assert len(out.pi_range) == 2, (
            "The `range` variable must be a tuple, " "with 2 values."
        )
        out.pi_range = tuple(out.pi_range)

        try:
            out.pi_center = kwargs["center"]
            out.pi_is_coded = False
        except KeyError:
            out.pi_center = np.mean(out.pi_range)

        try:
            out.pi_units = kwargs["units"]
            out.pi_is_coded = False
        except KeyError:
            out.pi_units = ""

        # Finally, the user might have over-ridden the coding flag:
        if override_coded is not None:
            out.pi_is_coded = override_coded

    else:
        if "levels" in kwargs:
            msg = "Levels must be list or tuple of the unique level names."
            # TODO: Check that all entries in the level list are accounted for.
            assert isinstance(kwargs.get("levels"), Iterable), msg
            out.pi_levels = {out.pi_name: list(kwargs.get("levels", []))}
        else:
            levels = out.unique()
            levels.sort()
            out.pi_levels = {out.pi_name: levels.tolist()}  # for use with Patsy

    units = kwargs.get("units", "")
    if units and not (out.pi_is_coded):
        out.name = f"{out.name} [{units}]"
    if out.pi_is_coded:
        out.name = f"{out.name} [coded]"

    return out


def expand_grid(**kwargs):
    """
    Create the expanded grid here.
    """
    n_col = len(kwargs)
    itrs = [v.values for v in kwargs.values()]
    product = list(itertools.product(*itrs))
    vals = np.fliplr(np.array(product).reshape(len(product), n_col))
    out = []
    for name, values in zip(kwargs.keys(), np.split(vals, n_col, axis=1)):
        out.append(c(values, name=name))

    return out


def supplement(x, **kwargs):
    return c(x.values, **kwargs)
    # (A, name = 'Feed rate', units='g/min', lo = 5, high = 8.0)
    # B = supplement(B, name = 'Initial inoculate amount', units = 'g', lo = 300,
    # hi = 400)
    # C = supplement(C, name = 'Feed substrate concentration', units = 'g/L',
    # lo = 40, hi = 60)
    # D = supplement(D, name = 'Dissolved oxygen set-point', units = 'mg/L',
    # lo = 4, hi = 5)


def gather(*args, title=None, **kwargs) -> Expt:
    """
    Gathers the named inputs together as columns for a data frame.

    Removes any rows that have ANY missing values. If even 1 value in a row
    is missing, then that row is removed.

    Usage
    -----

    expt = gather(A=A, B=B, y=y, title='My experiment in factors A and B')

    """
    # TODO : handle the case where the shape of an input >= 2 columns:  category

    out = Expt(data=None, index=None, columns=None, dtype=None)
    out.pi_source = defaultdict(str)
    out.pi_units = defaultdict(str)

    _ = [len(value) for value in kwargs.values()]
    index = []
    for key, value in kwargs.items():
        if isinstance(value, list):
            out[key] = value
        elif isinstance(value, pd.Series):
            out[key] = value.values
            out.pi_source[key] = value.name
            out.pi_units[key] = value.pi_units if hasattr(value, "pi_units") else ""

            if hasattr(value, "pi_index"):
                index.append(value.index)

        elif isinstance(value, pd.DataFrame):
            assert False, "Handle this case still"

    # TODO : check that all indexes are common, to merge. Or use the pandas
    #        functionality of merging series with the same index
    if index:
        out.index = index[0]

    # Drop any missing values:
    out.dropna(axis=0, how="any", inplace=True)

    # Set the title, if one was provided
    out.pi_title = title
    return out
